'use strict';
/**
 * A function that is fired when the pool is ready for it to execute.
 * It can return a promise if it is asynchronous, in which case the pool will only start another task once the promise resolves.
 * @callback Task
 * @returns {Promise|any}
 */

/**
 * Create a PromisePool
 */

function isNativeReflectConstruct() { if (typeof Reflect === "undefined" || !Reflect.construct) return false; if (Reflect.construct.sham) return false; if (typeof Proxy === "function") return true; try { Date.prototype.toString.call(Reflect.construct(Date, [], function () {})); return true; } catch (e) { return false; } }

function _construct(Parent, args, Class) { if (isNativeReflectConstruct()) { _construct = Reflect.construct; } else { _construct = function _construct(Parent, args, Class) { var a = [null]; a.push.apply(a, args); var Constructor = Function.bind.apply(Parent, a); var instance = new Constructor(); if (Class) _setPrototypeOf(instance, Class.prototype); return instance; }; } return _construct.apply(null, arguments); }

function _setPrototypeOf(o, p) { _setPrototypeOf = Object.setPrototypeOf || function _setPrototypeOf(o, p) { o.__proto__ = p; return o; }; return _setPrototypeOf(o, p); }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } }

function _createClass(Constructor, protoProps, staticProps) { if (protoProps) _defineProperties(Constructor.prototype, protoProps); if (staticProps) _defineProperties(Constructor, staticProps); return Constructor; }

var PromisePool =
/*#__PURE__*/
function () {
  _createClass(PromisePool, null, [{
    key: "create",

    /**
     * Create a new instance of the class with the specified arguments.
     * @param  {...any} args
     * @returns {PromisePool}
     */
    value: function create() {
      for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
        args[_key] = arguments[_key];
      }

      return _construct(this, args);
    }
    /**
     * Instantiate the PromisePool with the desired concurrency.
     * @param {Object} opts
     * @param {number} opts.concurrency - How many tasks to run at once.
     * @param {PromiseConstructor} [opts.PromiseClass=Promise] - The Promise class to use. It must support `Promise.resolve().finally(() => {})`. If you are using Node v10 or above, you don't have to modify this, as the default `Promise` class already supports `.finally`. An alternative to passing in a custom `PromiseClass`, is to polyfill the builtin `Promise` class.
     */

  }]);

  function PromisePool(_ref) {
    var concurrency = _ref.concurrency,
        _ref$PromiseClass = _ref.PromiseClass,
        PromiseClass = _ref$PromiseClass === void 0 ? Promise : _ref$PromiseClass;

    _classCallCheck(this, PromisePool);

    /**
     * How many tasks to run at once.
     * @type {number}
     * @private
     */
    this.concurrency = concurrency;
    /**
     * The Promise class to use.
     * It must support `Promise.resolve().finally(() => {})`.
     * @type {PromiseConstructor}
     * @private
     */

    this.PromiseClass = PromiseClass;
    /**
     * How many tasks are currently running.
     * @type {number}
     * @private
     */

    this.running = 0;
    /**
     * How many tasks have been started.
     * @type {number}
     * @private
     */

    this.started = 0;
    /**
     * For upcoming tasks, they have a function (a promise resolver) inserted into the queue which starts the task when called.
     * @type {Array<Function>}
     * @private
     */

    this.queue = [];
  }
  /**
   * Add a task to the pool.
   * @param {Task} task - The task to be executed when the pool permits.
   * @return {Promise} Returns a promise that resolves once the task has resolved or rejected. You should `.catch` it in case your task fails.
   */


  _createClass(PromisePool, [{
    key: "open",
    value: function open(task) {
      var _this = this;

      // Grab the class to use to create our promise,
      // so that the consumer can ensure `.finally` exists.
      var Promise = this.PromiseClass; // Create our promise and push its resolver to the queue.
      // This has the effect that we can queue its execution for later, instead of right now.

      var p = new Promise(function (resolve) {
        return _this.queue.push(resolve);
      }) // Once the resolver has fired, update the counts accordingly.
      .finally(function () {
        _this.started--;
        _this.running++;
      }) // Fire our our task and store the result.
      .then(task) // Update our counts accordingly, and start the next queue item if there are any.
      .finally(function () {
        _this.running--;

        if (_this.queue.length) {
          _this.started++;

          _this.queue.shift()();
        }
      }); // If our pool is under capacity, then start the first item in the queue.

      if (this.running + this.started < this.concurrency && this.queue.length) {
        this.started++;
        this.queue.shift()();
      } // Return the the promise that wraps the task,
      // such that it resolves once the task has compelted and our wrappers have completed.
      // This allows the user to do `Promise.all(Array<Task>.map((task) => pool.open(task)))`,
      // so that they can queue something for when all their pooled tasks are completed.
      // It also allows the user to do the mandatory `.catch` handling for task failures.


      return p;
    }
  }]);

  return PromisePool;
}();

module.exports = PromisePool;